package link;

import java.net.*;
import java.io.*;
import java.util.*;
import debug.*;
import timer.*;
import structure.*;

/**
  Implements a UDP Link. The Link establishes a UDP connection with 
  a remote node and actually sends datagram packets when the Transmit
  Packet method is called. Similarly it receives actual packets and 
  passes them on to an appropriate node. 
  */

public class UDPLink extends TimerClient
                     implements Link, Runnable
{
  /**
    Create a new link with the specified parameters
    
    @param name the debug name for this link
    
    @param mtu the maximum transmission unit, in bytes
    
    @param dest the hostname of the destintation
    @param port the port to use (for send and receive)
    
    @param bandwidthEst an estimate of the link bandwidth, in Mb/s
    @param latencyEst an estimate of the link latency, in ms
    */
  public UDPLink (String name, int mtu, String dest, int port,
		  double bandwidthEst, double latencyEst)
  {
    this.name = name;
    this.mtu = mtu;
    
    this.bandwidthEst = bandwidthEst;
    this.latencyEst = latencyEst;
    
    node = null;
    
    remotePort = port;
    
    try
      {
	if (dest.equals ("localhost"))
	  destAddr = InetAddress.getLocalHost ();
	else
	  destAddr = InetAddress.getByName (dest);
      }
    catch (Exception e)
      {
	System.out.println ("ERROR resolving hostname " + dest);
	System.exit (1);
      }
    
    try
      {
	socket = new DatagramSocket ();
      }
    catch (Exception e)
      {
	System.out.println ("ERROR opending UDP socket on port " + port);
	System.exit (1);
      }
    
    localPort = socket.getLocalPort ();
    
    Thread t = new Thread (this);
    t.start ();
  }

  public UDPLink (String name, int mtu, int port,
                   double bandwidthEst, double latencyEst)
  {
    this.name = name;
    this.mtu = mtu;
    
    this.bandwidthEst = bandwidthEst;
    this.latencyEst = latencyEst;
    
    node = null;
    
    localPort = port;
    remotePort = -1;
    destAddr = null;
    
    try
      {
	socket = new DatagramSocket (localPort);
      }
    catch (Exception e)
      {
	System.out.println ("ERROR opending UDP socket on port " + port);
	System.exit (1);
      }
    
    Thread t = new Thread (this);
    t.start ();
  }
  
  /**
    Convert to a string for printing
    */
  public String toString ()
  {
    return "UDP: " + name;
  }
  
  // The name of the link
  private String name;
  private int mtu;
  
  private double bandwidthEst;
  private double latencyEst;
  
  // The destination address, port and the local port
  private InetAddress destAddr;
  private int localPort;
  private int remotePort;
  private DatagramSocket socket;
  
  // The node locally attached to this link 
  private Node node;
  
  /**
    Transmit a packet on the link. 
    Actually sends the packet on the link. 
    
    @param p the packet to transmit
    @param n the node (router) sending the packet
    @see Link#TransmitPacket
    */
  public synchronized void TransmitPacket (LinkPacket p, Node n)
  {
    if (p.data.length > MTU ())
      return;
    
    StructuredStorage s = new StructuredStorage ();
    s.ConvertFromObject (p);
    
    byte[] data = s.ConvertToByteArray ();
    
    if (destAddr == null)
      return;
    
    DatagramPacket dp = new DatagramPacket (data, data.length,
					    destAddr, remotePort);
    
    try
      {
	if (socket != null)
	  socket.send (dp);
      }
    catch (IOException e)
      {
	System.out.println ("ERROR: UDP IOException sending datagram."+e);
	//System.exit (1);
      }
  }

  /**
    Get the maximum number of data bytes to transmit
    
    @return the maximum transmission unit, in bytes
    @see Link#MTU
    */
  public int MTU ()
  {
    return mtu;
  }
  
  public int getLocalPort() { 
    return localPort; 
  } 

  /**
    Get the average bandwidth for this link
    
    @return the bandwidth, in Mb/s
    @see Link#Bandwidth
    */
  public double Bandwidth ()
  {
    return bandwidthEst;
  }

  /**
    Get the average latency for this link
    
    @return the latency, in ms
    @see Link#Latency
    */
   public double Latency ()
   {
      return latencyEst;
   }
  
   public String[] OtherNodeNames (Node n)
   {
      return new String[0];
   }

   /**
     Add a node, maintains the list of connected nodes. 
     Only one node is connected to a UDP link 
     @see Link#AddNode
     */
  public void AddNode (Node n)
  {
    if (node == null)
      node = n;
  }

  /**
    Remove a node
    @see Link#RemoveNode
    */
  public void RemoveNode (Node n)
  {
    node = null;
  }
  
  /**
    Receives packets and calls the PacketArrived method of the node
    */
  public void run ()
  {
    for (;;)
      {
	byte[] data = new byte[MTU () * 2];
	DatagramPacket dp = new DatagramPacket (data, data.length);
	
	try
	  {
            socket.receive (dp);
	  }
	catch (IOException e)
	  {
            System.out.println ("ERROR: IOException receiving datagram."+e);
	    continue; 
            //System.exit (1);
	  }
	
	if (destAddr == null)
	  {
            destAddr = dp.getAddress ();
            remotePort = dp.getPort ();
	  }
	
	StructuredStorage s = new StructuredStorage ();
	s.ConvertFromByteArray (dp.getData ());
	LinkPacket p = (LinkPacket) s.ConvertToObject ();
	
	synchronized (Timer.global)
	  {
            if (node != null)
	      node.PacketArrived (p, this);
	  }
      }
  }
}
